package nl.iobyte.themeparkconnector.api.authentication.objects;

import org.bukkit.plugin.java.JavaPlugin;
import org.bukkit.scheduler.BukkitRunnable;
import java.util.Date;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

/**
 * A Weak Concurrent HashMap Solution which stores the keys and values only for a specific amount of time, and then expires after that
 * time.
 *
 * This HashMap uses the Bukkit scheduler for the cleanup.
 *
 * @author Vivekananthan M, Stijn Bannink
 *
 * @param <K> The key in the hashmap
 * @param <V> The value in the hashmap
 */
public class WeakConcurrentHashMap<K, V> extends ConcurrentHashMap<K, V> {

    private static final long serialVersionUID = 1L;

    private final Map<K, Long> timeMap = new ConcurrentHashMap<>();
    private final long expiryInMillis;

    private Listener<K, V> listener;
    private boolean mapAlive = true;

    public WeakConcurrentHashMap(JavaPlugin plugin) {
        this.expiryInMillis = 10000;
        initialize(plugin);
    }

    public WeakConcurrentHashMap(JavaPlugin plugin, Listener<K, V> listener) {
        this.listener = listener;
        this.expiryInMillis = 10000;
        initialize(plugin);
    }

    public WeakConcurrentHashMap(JavaPlugin plugin, long expiryInMillis) {
        this.expiryInMillis = expiryInMillis;
        initialize(plugin);
    }

    public WeakConcurrentHashMap(JavaPlugin plugin, long expiryInMillis, Listener<K, V> listener) {
        this.expiryInMillis = expiryInMillis;
        this.listener = listener;
        initialize(plugin);
    }

    private void initialize(JavaPlugin plugin) {
        new CleanerRunnable().runTaskTimerAsynchronously(plugin, expiryInMillis / 2, expiryInMillis / 2);
    }

    /**
     * {@inheritDoc}
     *
     * @throws IllegalStateException if trying to insert values into map after quiting
     */
    @Override
    public V put(K key, V value) {
        if (!mapAlive) {
            throw new IllegalStateException("WeakConcurrent Hashmap is no more alive.. Try creating a new one.");	// No I18N
        }
        Date date = new Date();
        timeMap.put(key, date.getTime());
        V returnVal = super.put(key, value);
        if (listener != null) {
            listener.notifyOnAdd(key, value);
        }
        return returnVal;
    }

    /**
     * {@inheritDoc}
     *
     * @throws IllegalStateException if trying to insert values into map after quiting
     */
    @Override
    public void putAll(Map<? extends K, ? extends V> m) {
        if (!mapAlive) {
            throw new IllegalStateException("WeakConcurrent Hashmap is no more alive.. Try creating a new one.");	// No I18N
        }
        for (K key : m.keySet()) {
            put(key, m.get(key));
        }
    }

    /**
     * {@inheritDoc}
     *
     * @throws IllegalStateException if trying to insert values into map after quiting
     */
    @Override
    public V putIfAbsent(K key, V value) {
        if (!mapAlive) {
            throw new IllegalStateException("WeakConcurrent Hashmap is no more alive.. Try creating a new one.");	// No I18N
        }
        if (!containsKey(key)) {
            return put(key, value);
        } else {
            return get(key);
        }
    }

    /**
     * Should call this method when it's no longer required
     */
    public void quitMap() {
        mapAlive = false;
    }

    public boolean isAlive() {
        return mapAlive;
    }

    /**
     *
     * This thread performs the cleaning operation on the concurrent hashmap once in a specified interval. This wait interval is half of the
     * time from the expiry time.
     *
     *
     */
    private class CleanerRunnable extends BukkitRunnable {
        @Override
        public void run() {
            if (mapAlive) cleanMap();
            else cancel();
        }

        private void cleanMap() {
            long currentTime = new Date().getTime();
            for (K key : timeMap.keySet()) {
                if (currentTime > (timeMap.get(key) + expiryInMillis)) {
                    V value = remove(key);
                    timeMap.remove(key);
                    if (listener != null) {
                        listener.notifyOnRemoval(key, value);
                    }
                }
            }
        }
    }

    public interface Listener<K,V> {
        void notifyOnAdd(K key, V value);
        void notifyOnRemoval(K key, V value);
    }

}
