package nl.iobyte.themeparkconnector.api.network;

import nl.iobyte.themepark.scheduler.ThemeParkScheduler;
import nl.iobyte.themeparkconnector.ThemeParkConnector;
import nl.iobyte.themeparkconnector.api.network.handlers.ClientConnectivityHandler;
import nl.iobyte.themeparkconnector.api.network.handlers.ClientOperatorHandler;
import nl.iobyte.themeparkconnector.api.network.io.SocketConnector;
import nl.iobyte.themeparkconnector.api.network.objects.INetworkingEvent;
import nl.iobyte.themeparkconnector.api.client.ClientConnection;
import nl.iobyte.themeparkconnector.api.packet.enums.PacketChannel;
import nl.iobyte.themeparkconnector.api.packet.objects.AbstractPacket;
import nl.iobyte.themeparkconnector.api.packet.objects.PayloadHandler;
import nl.iobyte.themeparkconnector.logger.ThemeParkConnectorLogger;
import org.bukkit.Bukkit;
import org.bukkit.entity.Player;
import java.util.*;

public class NetworkingService {

    private final Set<INetworkingEvent> eventHandlers = new HashSet<>();
    private final Map<UUID, ClientConnection> clientMap = new HashMap<>();
    private final Map<String, List<PayloadHandler<?>>> packetHandlerMap = new HashMap<>();
    private SocketConnector socketConnector;

    /**
     * Initialize SocketConnector
     */
    public void init() {
        if(socketConnector != null)
            return;

        socketConnector = new SocketConnector();

        //Connect, Disconnect, Kick
        ClientConnectivityHandler.load(this);

        //Operator
        ClientOperatorHandler.load(this);
    }

    /**
     * Establish a socket connection if the connection is down
     */
    public void connectIfDown() {
        ThemeParkScheduler.runAsync(() -> socketConnector.setupConnection());
    }

    public void send(ClientConnection client, AbstractPacket packet) {
        for (INetworkingEvent event : getEvents())
            event.onPacketSend(client, packet);

        socketConnector.send(client, packet);
    }

    /**
     * Get handler for packet and call it
     * @param abstractPacket Received Packet
     */
    public void triggerPacket(AbstractPacket abstractPacket) {
        if (packetHandlerMap.get(abstractPacket.getChannel()) == null) {
            ThemeParkConnectorLogger.toConsole("Unknown handler for packet channel " + abstractPacket.getChannel());
            return;
        }

        packetHandlerMap.get(abstractPacket.getChannel()).forEach(handler -> handler.handle(abstractPacket));
    }

    public void registerHandler(PacketChannel channel, PayloadHandler<?> handler) {
        registerHandler(channel.toString(), handler);
    }

    /**
     * Link a handler to a packet type
     * @param channel Channel of the packet
     * @param handler Handler for the incoming data
     */
    public void registerHandler(String channel, PayloadHandler<?> handler) {
        List<PayloadHandler<?>> handlers = packetHandlerMap.getOrDefault(channel, new ArrayList<>());
        handlers.add(handler);
        packetHandlerMap.put(channel, handlers);
    }

    /**
     * Get Client from UUID
     * @param uuid UUID of client
     * @return ClientConnection instance
     */
    public ClientConnection getClient(UUID uuid) {
        if (clientMap.containsKey(uuid))
            return clientMap.get(uuid);

        Player player = Bukkit.getPlayer(uuid);
        if (player == null)
            return null;

        return register(player);
    }

    /**
     * Returns all ClientConnections
     * @return Collection of all ClientConnections
     */
    public Collection<ClientConnection> getClients() {
        return clientMap.values();
    }

    /**
     * Remove client from the NetworkingService
     * @param player Player's UUID
     */
    public void remove(UUID player) {
        if (!clientMap.containsKey(player))
            return;

        clientMap.remove(player).remove();
    }

    /**
     * Register client to the NetworkingService
     * @param player Player
     * @return ClientConnection instance
     */
    public ClientConnection register(Player player) {
        ClientConnection clientConnection = new ClientConnection(player);
        clientMap.put(player.getUniqueId(), clientConnection);
        return clientConnection;
    }

    /**
     * Force the socket connection to close
     */
    public void stop() {
        socketConnector.disconnect();
    }

    /**
     * Get networking events
     * @return Set of NetworkingEvents
     */
    public Set<INetworkingEvent> getEvents() {
        return eventHandlers;
    }

    /**
     * Add event handler for networking activity
     * @param events Networking event
     */
    public void addEventHandler(INetworkingEvent events) {
        eventHandlers.add(events);
    }

}
